#pragma once

/** \file LinearSolverInterface.h */

/**
 * \defgroup Solver Solver module
 * @{ <img src="solver.png"> @}
 *
 * \brief
 * Contains all classes needed for solving linear and non linear equation
 * systems.
 */

#include <amdis/common/TypeTraits.hpp>

#include <amdis/linearalgebra/PreconditionerInterface.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>

namespace AMDiS
{
  /// Abstract base class for linear solvers
  template <class Traits>
  class LinearSolverInterface
  {
  protected:
    using RunnerBase = RunnerInterface<Traits>;
    using Mat = typename Traits::Mat;
    using Sol = typename Traits::Sol;
    using Rhs = typename Traits::Rhs;
    using Comm = typename Traits::Comm;

  public:
    /// Destructor.
    virtual ~LinearSolverInterface() = default;

    /// Public method to call in order to solve a linear system Ax = b.
    /**
     * The method redirects to the specific linear solver and prints statistics
     * and error estimations at the end.
     *
     * The parameters correspond to
     *  \p A     A [block-]matrix that represents the system-matrix.
     *  \p x     A [block-]vector for the unknown components.
     *  \p b     A [block-]vector for the right-hand side of the linear system.
     **/
    void solve(Mat const& A, Sol& x, Rhs const& b, Comm& comm, SolverInfo& solverInfo)
    {
      solveImpl(A, x, b, comm, solverInfo);
    }

    // return the runner/worker corresponding to this solver (optional)
    virtual std::shared_ptr<RunnerBase> runner() { return {}; };

  private:
    /// main methods that all solvers must implement
    virtual void solveImpl(Mat const& A, Sol& x, Rhs const& b, Comm& comm, SolverInfo& solverInfo) = 0;
  };


  template <class Matrix, class Vector>
  struct SolverCreator;

} // end namespace AMDiS
