#pragma once

#include <memory>
#include <ostream>
#include <set>
#include <string>
#include <utility>

#include <dune/common/version.hh>
#include <dune/istl/owneroverlapcopy.hh>
#include <dune/istl/solvercategory.hh>

#include <amdis/functions/GlobalIdSet.hpp>

namespace AMDiS
{

#if HAVE_MPI
  /// Implementation class for ISTL-specific communication to be used in parallel solvers
  template <class Basis>
  class ISTLCommunication
      : public Dune::OwnerOverlapCopyCommunication<typename GlobalBasisIdSet<Basis>::IdType, typename Basis::size_type>
  {
    using GlobalCommIdType = typename GlobalBasisIdSet<Basis>::IdType;
    using LocalCommIdType = typename Basis::size_type;

    using Self = ISTLCommunication;
    using Super = Dune::OwnerOverlapCopyCommunication<GlobalCommIdType, LocalCommIdType>;

    using GridView = typename Basis::GridView;
    using Grid = typename GridView::Grid;
    using Element = typename GridView::template Codim<0>::Entity;
    using IdSet = typename Grid::GlobalIdSet;
    using IdType = typename IdSet::IdType;
    using EntitySet = std::set<IdType>;
    using Comm = typename Dune::MPIHelper::MPICommunicator;
    using SolverType = typename Dune::SolverCategory::Category;

  public:
    /**
     * Constructor. Takes the global basis, an MPI communicator and the ISTL solver type
     *
     * \param basis   Global basis associated to the solvers' matrices and vectors
     * \param comm    MPI communicator of the group containing the distributed matrices and vectors
     * \param cat     Category of the solver used with this communicator, one of
     *                Dune::SolverCategory::Category::sequential, overlapping, nonoverlapping
     **/
    ISTLCommunication(Basis const& basis, Comm const& comm, SolverType cat)
      : Super(comm, cat)
    {
      if (cat != SolverType::sequential)
      {
        auto& pis = this->indexSet();
        buildParallelIndexSet(basis, pis);
        auto& ris = this->remoteIndices();
        ris.setIndexSets(pis,pis,comm);
        ris.template rebuild<true>();
      }
    }

  private:
    /// Builds the parallel index set containing all DoFs of the basis
    void buildParallelIndexSet(Basis const& basis, typename Super::PIS& pis);

  public:
    /**
     * ISTLCommunication creator function. Takes the global basis and the solver's initfile prefix
     * and returns a unique pointer to the ISTLCommunication object.
     * Deduces the category of the solver used from the initfile and basis
     *
     * \param basis   Global basis associated to the solvers' matrices and vectors
     * \param prefix  Solver prefix used in the initfile, usually "<problem name>->solver"
     **/
    static std::unique_ptr<Self> create(Basis const& basis, std::string const& prefix);
  };

#else // !HAVE_MPI

  /// Dummy implementation for ISTL-specific communication when no MPI is found
  template <class Basis>
  class ISTLCommunication
      : public DefaultCommunication<Basis>
  {
    using Self = ISTLCommunication;

  public:
    ISTLCommunication() = default;

    Dune::SolverCategory::Category category() const
    {
      return Dune::SolverCategory::Category::sequential;
    }

    static std::unique_ptr<Self> create(Basis const& basis, std::string const& prefix)
    {
      DUNE_UNUSED_PARAMETER(basis);
      DUNE_UNUSED_PARAMETER(prefix);
      return std::make_unique<Self>();
    }
  };
#endif

} // end namespace AMDiS

#include <amdis/linearalgebra/istl/Communication.inc.hpp>
