#pragma once

#include <algorithm>
#include <string>

#include <dune/common/ftraits.hh>

#include <amdis/linearalgebra/istl/Fwd.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/Output.hpp>

namespace AMDiS
{
  /**
  * \ingroup Solver
  * \class AMDiS::DirectRunner
  * \brief \implements RunnerInterface for the (external) direct solvers
  */
  template <template <class> class Solver, class Traits>
  class DirectRunner
      : public RunnerInterface<Traits>
  {
  protected:
    using Super = RunnerInterface<Traits>;
    using Mat = typename Traits::Mat;
    using Sol = typename Traits::Sol;
    using Rhs = typename Traits::Rhs;
    using Comm = typename Traits::Comm;
    using BaseSolver  = Dune::InverseOperator<Sol, Rhs>;

  public:
    /// Constructor.
    DirectRunner(std::string const& prefix)
      : solverCreator_(prefix)
    {}

    /// Implements \ref RunnerInterface::init()
    void init(Mat const& A, Comm& comm) override
    {
      DUNE_UNUSED_PARAMETER(comm);
      solver_ = solverCreator_.create(A);
    }

    /// Implements \ref RunnerInterface::exit()
    void exit() override
    {
      solver_.reset();
    }

    /// Implements \ref RunnerInterface::solve()
    int solve(Mat const& A, Sol& x, Rhs const& b, SolverInfo& solverInfo) override
    {
      DUNE_UNUSED_PARAMETER(A);

      // storing some statistics
      Dune::InverseOperatorResult statistics;

      // solve the linear system
      Rhs _b = b;
      solver_->apply(x, _b, statistics);

      solverInfo.setRelResidual(statistics.reduction);

      return 0;
    }

  private:
    ISTLSolverCreator<Solver<Mat>> solverCreator_;
    std::shared_ptr<BaseSolver> solver_;
  };
}
